import os
import bisect
import itertools

from enum import IntEnum
from typing import Type, Any, Tuple, Dict, List, Optional, Callable
from pydantic import BaseModel, Field, FilePath, DirectoryPath, PositiveInt, NonNegativeInt, field_validator, model_validator

from video_dataset.padder import Padder
from video_dataset.utils import better_listdir
from video_dataset.video import Video, UndefinedVideoException
from video_dataset.annotations import Annotations, UndefinedAnnotationsException

class VideoShapeComponents(IntEnum):
    TIME = 0
    HEIGHT = 1
    WIDTH = 2
    CHANNELS = 3
    
DEFAULT_VIDEO_SHAPE = (VideoShapeComponents.TIME, VideoShapeComponents.HEIGHT, VideoShapeComponents.WIDTH, VideoShapeComponents.CHANNELS)

class VideoDatasetConfiguration(BaseModel):
    annotations_dir: DirectoryPath
    videos_dir: DirectoryPath
    video_processor: Type[Video]
    annotations_processor: Type[Annotations]
    segment_size: int
    verbose: bool = True
    video_extension: str = 'mp4'
    annotations_extension: str = 'csv'
    video_shape: Tuple[NonNegativeInt, NonNegativeInt, NonNegativeInt, NonNegativeInt] = Field(default=DEFAULT_VIDEO_SHAPE)
    step: Optional[PositiveInt] = 1
    
    video_processor_kwargs: Optional[Dict[str, Any]] = {}
    annotations_processor_kwargs: Optional[Dict[str, Any]] = {}
    ids_file: Optional[FilePath] = None
    frames_transform: Optional[Callable] = None
    annotations_transform: Optional[Callable] = None
    
    allow_undefined_annotations: bool = False
    
    padder: Optional[Any] = None
    
    overlap: Optional[NonNegativeInt] = 0
    
    load_videos: Optional[bool] = True
    load_annotations: Optional[bool] = True

    @model_validator(mode='before')
    @classmethod
    def adjust_loading_videos_on_segment_size(cls, values):
        segment_size = values.get('segment_size')
        
        load_videos_set = 'load_videos' in values
        
        # NOTE: if segment_size is -1 and loading flags are not explicitly set, set them to False just to prevent unnecessary unwanted loading times.
        if segment_size == VideoDataset.FULL_VIDEO_SEGMENT and not load_videos_set:
            values['load_videos'] = False
        
        return values

    @field_validator("segment_size")
    def check_segment_size(cls, v):
        if v < VideoDataset.FULL_VIDEO_SEGMENT:
            raise ValueError("segment_size must be bigger or equal to -1.")
        return v

    @field_validator("video_processor")
    def check_video_processor(cls, v):
        if not issubclass(v, Video):
            raise ValueError("Video processor must inherit from Video class.")
        return v

    @field_validator("annotations_processor")
    def check_annotations_processor(cls, v):
        if not issubclass(v, Annotations):
            raise ValueError("Annotations processor must inherit from Annotations class.")
        return v

    @field_validator("video_shape")
    def check_video_shape(cls, v):
        if not isinstance(v, tuple) or len(v) != 4 or len(set(v)) != 4:
            raise ValueError("Video shape must have exactly 4 unique components.")
        return v

    @field_validator("video_processor_kwargs", "annotations_processor_kwargs", mode="before")
    def check_kwargs(cls, v):
        if v is not None and not isinstance(v, dict):
            raise ValueError("Processor kwargs must be a dictionary or None.")
        return v

    @field_validator("frames_transform", "annotations_transform", mode="before")
    def check_transform_callable(cls, v):
        if v is not None and not callable(v):
            raise ValueError("Transform must be callable or None.")
        return v
    
    # @field_validator("padder")
    # def check_padder(cls, v):
    #     if v is not None and not isinstance(v, Padder):
    #         raise ValueError("Padder must be an instance of a subclass of Padder.")
    #     return v
    
class VideoDataset():
    FULL_VIDEO_SEGMENT = -1
    
    def __init__(self, **kwargs):
        configuration = VideoDatasetConfiguration(**kwargs)
        
        self.__dict__.update(configuration.model_dump())

        if self.ids_file is None:
            self.ids = list(map(lambda file_name: os.path.splitext(file_name)[0], better_listdir(self.videos_dir)))
        else:
            with open(self.ids_file, "r") as file:
                self.ids = file.read().splitlines()
                
        self.videos, self.annotations = self.__prepare_videos_and_annotations()
        
        self.__segment_size_check()
        
    def __prepare_videos_and_annotations(self):
        videos = []
        annotations = []
        
        for id in self.ids:
            video = self.video_processor(self.videos_dir, id, **self.video_processor_kwargs)
                
            try:
                annotation = self.annotations_processor(self.annotations_dir, id, **self.annotations_processor_kwargs)
            except UndefinedAnnotationsException as exception:
                if self.allow_undefined_annotations:
                    if self.verbose:
                        print(f"[warning]: {exception.message}")
                    annotation = None
                else:
                    raise exception
            
            videos.append(video)
            annotations.append(annotation)
            
        return videos, annotations
        
    def __segment_size_check(self):
        if self.padder is None and self.segment_size != VideoDataset.FULL_VIDEO_SEGMENT:
            for index, video in enumerate(self.videos):
                remaining_segments = len(video) % self.segment_size
                if remaining_segments != 0:
                    if self.verbose:
                        print(f"[warning]: {remaining_segments} frames will be lost, because video {index} has {len(video)} frames, which is not divisible by segment size {self.segment_size}. consider using a padder.")

    def __len__(self):
        if self.segment_size == VideoDataset.FULL_VIDEO_SEGMENT:
            return len(self.videos)
        else:
            return sum([(max(0, len(video) - self.overlap) // (self.segment_size - self.overlap)) for video in self.videos])
        
    def __getitem__(self, virtual_video_index):
        if self.segment_size == VideoDataset.FULL_VIDEO_SEGMENT:
            video_index = virtual_video_index
            
            frames = self.__getitem_frames__(video_index, 0) if self.load_videos else None
            annotations = self.__getitem_annotations__(video_index, 0) if self.load_annotations else None
            
            return frames, annotations, (video_index, self.videos[video_index].get_id(), 0)
        else:
            video_index, starting_frame_number_in_video = self.__translate_virtual_video_index_to_video_index(virtual_video_index)
            
            frames = self.__getitem_frames__(video_index, starting_frame_number_in_video) if self.load_videos else None
            annotations = self.__getitem_annotations__(video_index, starting_frame_number_in_video) if self.load_annotations else None
        
            return frames, annotations, (video_index, self.videos[video_index].get_id(), starting_frame_number_in_video)
    
    def __getitem_frames__(self, video_index, starting_frame_number_in_video):
        if self.segment_size == VideoDataset.FULL_VIDEO_SEGMENT:
            frames = self.videos[video_index][0:]
            
            frames = frames.transpose(self.video_shape)
            
            if self.frames_transform is not None:
                frames = self.frames_transform(frames)
                
            return frames
        else:
            starting_frame = starting_frame_number_in_video
            ending_frame = starting_frame_number_in_video + self.segment_size
            
            frames = self.videos[video_index][starting_frame:ending_frame:self.step]
            
            # NOTE: we expect the video_processor to return a numpy array of the frames in the DEFAULT_VIDEO_SHAPE format.
            frames = frames.transpose(self.video_shape)
            
            if self.padder is not None:
                frames, _ = self.padder(frames=frames, annotations=None,  target_segment_size=self.segment_size // self.step)
            
            if self.frames_transform is not None:
                frames = self.frames_transform(frames)
            
            return frames
    
    def __getitem_annotations__(self, video_index, starting_frame_number_in_video):
        if self.segment_size == VideoDataset.FULL_VIDEO_SEGMENT:
            video_annotations = self.annotations[video_index]
            
            if video_annotations is None and self.allow_undefined_annotations:
                return None
            
            else:
                annotations = video_annotations[0:]
                
                if self.annotations_transform is not None:
                    annotations = self.annotations_transform(annotations)
                    
                return annotations
        else:
            starting_frame = starting_frame_number_in_video
            ending_frame = starting_frame_number_in_video + self.segment_size
            
            video_annotations = self.annotations[video_index]
            
            if video_annotations is None and self.allow_undefined_annotations:
                return None
            else:
                annotations = video_annotations[starting_frame:ending_frame:self.step]
                
                if self.padder is not None:
                    _, annotations = self.padder(frames=None, annotations=annotations,  target_segment_size=self.segment_size // self.step)
                
                if self.annotations_transform is not None:
                    annotations = self.annotations_transform(annotations)
                    
                return annotations
    
    def __translate_virtual_video_index_to_video_index(self, virtual_video_index):
        video_index = 0
        
        videos_cropped_frames_number = [len(video) // self.segment_size for video in self.videos]
        
        cumulative_videos_cropped_frames_number = list(itertools.accumulate(videos_cropped_frames_number))
        
        video_index = bisect.bisect_right(cumulative_videos_cropped_frames_number, virtual_video_index)
        
        if video_index >= len(videos_cropped_frames_number):
            raise ValueError("Virtual video index is out of range.")

        previous_frames = 0 if video_index == 0 else cumulative_videos_cropped_frames_number[video_index - 1]
        segment_index_within_video = virtual_video_index - previous_frames
        starting_frame_number_in_video = segment_index_within_video * (self.segment_size - self.overlap)
        
        return video_index, starting_frame_number_in_video


import numpy as np

from typing import List, Any
from abc import ABC, ABCMeta, abstractmethod

class Padder(ABC):
    __metaclass__ = ABCMeta
    
    @abstractmethod
    def __init__(self):
        pass
    
    @abstractmethod
    def __call__(self, frames: np.ndarray = None, annotations: List[Any] = None, target_segment_size: int = None):
        pass
    
class ValuePadder(Padder):
    def __init__(self, frames_padding_value: Any, annotations_padding_value: Any):
        self.frames_padding_value = frames_padding_value
        self.annotations_padding_value = annotations_padding_value
        
    def __call__(self, frames: np.ndarray = None, annotations: List[Any] = None, target_segment_size: int = None):
        assert target_segment_size != None
        
        frames_padded = self.__pad_frames(frames, target_segment_size) if frames is not None else None
        annotations_padded = self.__pad_annotations(annotations, target_segment_size) if annotations is not None else None
        
        return frames_padded, annotations_padded
    
    def __pad_frames(self, frames: np.ndarray, target_segment_size: int):
        frames_padded = np.full((target_segment_size, *frames.shape[1:]), self.frames_padding_value, dtype=frames.dtype)
        frames_padded[:frames.shape[0]] = frames
        
        return frames_padded
    
    def __pad_annotations(self, annotations: List[Any], target_segment_size: int):
        annotations_padded = [self.annotations_padding_value] * target_segment_size
        annotations_padded[:len(annotations)] = annotations
        
        return annotations_padded
    
class LastValuePadder(Padder):
    def __init__(self):
        pass
    
    def __call__(self, frames: np.ndarray = None, annotations: List[Any] = None, target_segment_size: int = None):
        assert target_segment_size != None
        
        frames_padded = self.__pad_frames(frames, target_segment_size) if frames is not None else None
        annotations_padded = self.__pad_annotations(annotations, target_segment_size) if annotations is not None else None
        
        return frames_padded, annotations_padded
    
    def __pad_frames(self, frames: np.ndarray, target_segment_size: int):
        pad_length = target_segment_size - frames.shape[0]
        
        if pad_length <= 0:
            return frames  # No need to pad
        
        last_frame = frames[-1:]  # Correctly extract the last frame
        frames_padded = np.vstack([frames] + [last_frame] * pad_length)  # Stack repeated frames

        return frames_padded
    
    def __pad_annotations(self, annotations: List[Any], target_segment_size: int):
        annotations_padded = [annotations[-1]] * (target_segment_size - len(annotations))
        annotations_padded = annotations + annotations_padded
        
        return annotations_padded


import os
import csv

from typing import Union
from abc import ABC, ABCMeta, abstractmethod

class Annotations(ABC):
    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, annotations_dir_path: str, id: str):
        pass
    
    @abstractmethod
    def get_id(self):
        pass
    
    @abstractmethod
    def __getitem__(self, index: int | slice):
        """
        Get the annotation(s) of the video file corresponding to the given frame(s) index / indices.
        Note that even if an index is given the annotations will be returned in a batch format (Number of frames, Height, Width, Channels).
        """
        pass
    
class UndefinedAnnotationsException(Exception):
    """
    Raised when no valid annotations file is found for a given video ID.
    """
    
    def __init__(self, message: str, id: str):
        super().__init__(message)
        
        self.message = message
        self.id = id
    
FALLBACK_ANNOTATION = "nothing"
ALLOWED_INDEX_OVERFLOW = 100
    
class AnnotationsFromFrameLevelTxtFileAnnotations(Annotations):
    def __init__(self, annotations_dir_path, id, fallback_annotation = FALLBACK_ANNOTATION, max_overflow_value = ALLOWED_INDEX_OVERFLOW):
        self.annotations_dir_path = annotations_dir_path
        self.id = id
        self.fallback_annotation = fallback_annotation
        self.max_overflow_value = max_overflow_value
        
        if not self.__does_annotations_file_exists():
            raise UndefinedAnnotationsException(f"No valid annotations file found for {id} in {annotations_dir_path}", self.id)
        
    def __does_annotations_file_exists(self):
        return os.path.exists(os.path.join(self.annotations_dir_path, f"{self.id}.txt"))
        
    def get_id(self):
        return self.id
    
    def __len__(self):
        with open(os.path.join(self.annotations_dir_path, f"{self.id}.txt"), 'r') as f:
            lines = f.readlines()
            
        return len(lines)
    
    def __getitem__(self, index: int | slice):
        if isinstance(index, int):
            if index < 0 or index >= self.__len__() + self.max_overflow_value:
                raise IndexError("Index out of bounds")
            return self.__get_annotation(index)
        elif isinstance(index, slice):
            start, stop, step = index.indices(self.__len__() + self.max_overflow_value)
            return self.__get_annotations(start, stop, step)
        else:
            raise TypeError("Index must be an integer or slice")
        
    def __get_annotation(self, index: int):
        with open(os.path.join(self.annotations_dir_path, f"{self.id}.txt"), 'r') as f:
            lines = f.readlines()
        
        if index >= len(lines):
            return self.fallback_annotation
            
        # each line is an annotation for a frame
        return lines[index]
        
    def __get_annotations(self, start: int, stop: int, step: int):
        with open(os.path.join(self.annotations_dir_path, f"{self.id}.txt"), 'r') as f:
            lines = f.readlines()
            
        # each line is an annotation for a frame
        
        annotations = []
        for i in range(start, stop, step):
            if i >= len(lines):
                annotations.append(self.fallback_annotation)
            else:
                annotations.append(lines[i])
        
        return annotations
    
DEFAULT_CSV_DELIMITER = ';'

class AnnotationsFromSegmentLevelCsvFileAnnotations(Annotations):
    def __init__(self, annotations_dir_path: str, id: str, fps: int, fallback_annotation=FALLBACK_ANNOTATION, max_overflow_value=ALLOWED_INDEX_OVERFLOW, delimiter=DEFAULT_CSV_DELIMITER):
        self.annotations_dir_path = annotations_dir_path
        self.id = id
        self.fps = fps
        self.fallback_annotation = fallback_annotation
        self.max_overflow_value = max_overflow_value
        self.delimiter = delimiter
        
        if not self.__does_annotations_file_exists():
            raise UndefinedAnnotationsException(f"No valid annotations file found for {id} in {annotations_dir_path}", self.id)
        
        self.annotations = self.__load_annotations()
        
    def __does_annotations_file_exists(self):
        return os.path.exists(os.path.join(self.annotations_dir_path, f"{self.id}.csv"))
        
    def __load_annotations(self):
        """
        Given a csv file with annotations for segments, load the annotations into a list of frame level annotations.
        """
        annotations = []
        max_frame = 0
        
        with open(os.path.join(self.annotations_dir_path, f"{self.id}.csv"), newline='') as file:
            csv_reader = csv.DictReader(file, delimiter=self.delimiter)
            
            for row in csv_reader:
                start_frame = int(float(row['starting-timestamp']) * self.fps / 1000)
                end_frame = int(float(row['ending-timestamp']) * self.fps / 1000)
                
                if end_frame > max_frame:
                    max_frame = end_frame
                
                while len(annotations) < start_frame:
                    annotations.append(self.fallback_annotation)
                
                annotations.extend([row['action']] * (end_frame - start_frame + 1))
            
            return annotations
        
    def get_id(self):
        return self.id
    
    def __len__(self):
        return len(self.annotations)
    
    def __getitem__(self, index: Union[int, slice]):
        if isinstance(index, int):
            if index < 0 or index >= len(self) + self.max_overflow_value:
                raise IndexError("Index out of bounds")
            return self.__get_annotation(index)
        elif isinstance(index, slice):
            start, stop, step = index.indices(len(self) + self.max_overflow_value)
            return self.__get_annotations(start, stop, step)
        else:
            raise TypeError("Index must be an integer or slice")
        
    def __get_annotation(self, index: int):
        if index >= len(self.annotations):
            return self.fallback_annotation
        return self.annotations[index]
    
    def __get_annotations(self, start: int, stop: int, step: int):
        return [self.__get_annotation(i) for i in range(start, stop, step)]